import requests
from email.mime.multipart import MIMEMultipart
from email.mime.application import MIMEApplication
import email.encoders
from lxml import etree

from .exception import make_sendpeppol_error
from .wsse import encrypt_as4_document
from .wsse import sign_as4_envelope_and_body
from .wsse import insert_encryption_info_in_as4_envelope
from .as4 import generate_as4_envelope

def get_headers_and_body_for_posting_as4_document(document_content, document_xml, utc_timestamp, document_type, process_type, sender_id, receiver_id, to_party_id, xmlsec_path, keyfile, keyfile_password, sender_cert, receiver_cert, service_provider_id):
    message, gzip, attachment_id = make_as4_message_to_post(utc_timestamp, document_type, process_type, sender_id, receiver_id, to_party_id, document_content, xmlsec_path, keyfile, keyfile_password, sender_cert, receiver_cert, service_provider_id)

    related = MIMEMultipart('related')

    mt = MIMEApplication(message, 'soap+xml', email.encoders.encode_7or8bit)
    mt.add_header("Content-ID", "<root.message@cxf.apache.org>")
    related.attach(mt)

    mt = MIMEApplication(gzip, 'octet-stream')
    mt.add_header("Content-ID", f'<{attachment_id[4:]}>')
    related.attach(mt)

    # phase4 needs CRLF, oxalis is fine...
    body = bytes(related).replace(b'\n', b'\r\n')
    headers = dict(related.items())

    return body, headers

def post_edelivery_as4_document(endpoint_url, body, headers, timeout):
    try:
        r = requests.post(endpoint_url, data=body, headers=headers, timeout=timeout)
    except (ConnectionError, requests.exceptions.RequestException) as e:
        raise make_sendpeppol_error(str(e), 'server-error', temporary=True, url=endpoint_url)

    try:
        response_xml = etree.fromstring(r.content)
    except (etree.XMLSyntaxError, ValueError) as e:
        raise make_sendpeppol_error(str(e), 'server-error', temporary=True, url=endpoint_url)

    if not r.status_code == 200 or response_xml.find('.//{*}Header//{*}Receipt') is None:
        error_node = response_xml.find('.//{*}Error')
        error_msg = None
        if error_node is not None:
            error_msg = error_node.findtext('{*}ErrorDetail')

        exc_text = "Endpoint error"
        if error_msg:
            exc_text += ': ' + error_msg

        e = make_sendpeppol_error(
            exc_text, 'server-error', temporary=True, url=endpoint_url,
            endpoint={
                'status_code': r.status_code,
                'error_xml': etree.tostring(error_node if error_node is not None else response_xml, pretty_print=True).decode(),
                'error_msg': error_msg,
            })

        raise e

def make_as4_message_to_post(utc_timestamp, document_type, process_type, sender_id, receiver_id, to_party_id, document_content, xmlsec_path, keyfile, password, sender_cert, receiver_cert, service_provider_id):

    cipher_value, encrypted_gzipped_content, document_hash = encrypt_as4_document(document_content, receiver_cert, xmlsec_path)

    attachment_id, envelope, messaging, body = generate_as4_envelope(utc_timestamp, document_type, process_type, sender_id, receiver_id, to_party_id, service_provider_id)

    sign_as4_envelope_and_body(envelope, attachment_id, document_hash, body, messaging, sender_cert, keyfile, password)

    insert_encryption_info_in_as4_envelope(envelope, receiver_cert, cipher_value, attachment_id)

    message = etree.tostring(envelope)

    #print("AS4 message:", etree.tostring(envelope, pretty_print=True).decode('utf-8'))

    return message, encrypted_gzipped_content, attachment_id



import argparse
import sys
from pathlib import Path

from lxml import etree

from . import validate_peppol_document, send_peppol_document, SendPeppolError


def enable_debug_logging():
    import http.client as http_client
    http_client.HTTPConnection.debuglevel = 1

    import logging
    logging.basicConfig()
    logging.getLogger().setLevel(logging.DEBUG)
    requests_log = logging.getLogger("requests.packages.urllib3")
    requests_log.setLevel(logging.DEBUG)
    requests_log.propagate = True


def main():
    parser = argparse.ArgumentParser(description="Peppol utility", prog="python -m peppol_py")
    parser.add_argument('--verbose', action=argparse.BooleanOptionalAction, help="Enable debug logging")

    subparsers = parser.add_subparsers(help="command", dest="command")

    parser_send = subparsers.add_parser('send', help='Send a Peppol document')
    parser_send.add_argument('--document', help="The path of the document to send", required=True)
    parser_send.add_argument('--xmlsec-path', default='xmlsec1', help="The path to latest xmlsec binary")
    parser_send.add_argument('--schematron-path', default=['PEPPOL-EN16931-UBL.xsl'],
                             nargs='+', help="Schematron XSL files to validate with, defaults to bundled PEPPOL-EN16931-UBL.xsl")
    parser_send.add_argument('--keyfile', default='key.pem', help="The path to the private key, defaults to key.pem")
    parser_send.add_argument('--password', default='', help="The password for the private key, defaults to empty")
    parser_send.add_argument('--certfile', default='cert.pem', help="The path to the public key, defaults to cert.pem")
    parser_send.add_argument('--unwrap-sbh', action=argparse.BooleanOptionalAction,
                             help="Unwrap standard business header already present in document. Useful for testbed.")
    parser_send.add_argument('--test', action=argparse.BooleanOptionalAction, help="Use test SMP server, defaults to production")

    parser_validate = subparsers.add_parser('validate', help='Validate a Peppol document')
    parser_validate.add_argument('--document', help="The path of the document to validate", required=True)
    parser_validate.add_argument('--schematron-path', default=['PEPPOL-EN16931-UBL.xsl'],
                             nargs='+', help="Schematron XSL files to validate with, defaults to bundled PEPPOL-EN16931-UBL.xsl")

    parsed_args = parser.parse_args()

    if parsed_args.verbose:
        enable_debug_logging()

    if parsed_args.command == "send":
        with open(parsed_args.document, 'rb') as f:
            document_content = f.read()

        errors = validate_peppol_document(document_content, parsed_args.schematron_path)
        if errors:
            for d in errors:
                print(d)
            sys.exit(1)

        if parsed_args.unwrap_sbh:
            document_content = etree.tostring(etree.fromstring(document_content).find('./{*}Invoice'))

        try:
            stats = send_peppol_document(document_content,
                                         parsed_args.xmlsec_path, parsed_args.keyfile,
                                         keyfile_password=parsed_args.password, certfile=parsed_args.certfile,
                                         test_environment=parsed_args.test, document_type_version='2.1')
            print(stats)
        except SendPeppolError as ex:
            print(f"Failed with: {ex.code} {ex}")
            sys.exit(1)
        except Exception as ex:
            print(f"Failed with: {ex}")
            sys.exit(1)
    elif parsed_args.command == "validate":
        with open(parsed_args.document, 'rb') as f:
            document_content = f.read()

        errors = validate_peppol_document(document_content, parsed_args.schematron_path)
        if errors:
            for d in errors:
                print(d)
            sys.exit(1)
    else:
        pass


if __name__ == "__main__":
    main()



import os
from pathlib import Path
from typing import List

from lxml import etree
from saxonche import PySaxonProcessor


def validate_peppol_document(
    document_content: bytes,
    schematron_xsls: List[str],
    remove_namespaces_from_errors=True,
    warnings=False
) -> List[dict]:
    """
    Validate peppol document. Returns a list of errors.

    ``document_content`` (bytes): XML content

    ``schematron_xsls`` (list of str): Either full paths to xsl files, or names of xsl files shipped with peppol_py.

    ``remove_namespaces_from_errors`` (bool): For shorter output of error messages.

    ``warnings`` (bool): If ``True``, warnings will be returned as well out.
    """
    errors = []
    # we need saxonche (SaxonC Home Edition) because lxml (libxsl)
    # only works with XSLT 1.0, and the Schematron is written in XSLT
    # 2.0
    with PySaxonProcessor(license=False) as proc:
        for validation_xsl_file in schematron_xsls:
            if not os.path.exists(validation_xsl_file):
                # Resolve internal files
                validation_xsl_file = str(Path(__file__).parent / 'data' / 'sendpeppol-schematron' / validation_xsl_file)
            xsltproc = proc.new_xslt30_processor()
            document = proc.parse_xml(xml_text=document_content.decode())
            executable = xsltproc.compile_stylesheet(stylesheet_file=validation_xsl_file)
            output = executable.transform_to_string(xdm_node=document)

            parsed_output = etree.fromstring(output.encode())
            for e in parsed_output.findall('{http://purl.oclc.org/dsdl/svrl}failed-assert'):
                location = e.get('location')
                if remove_namespaces_from_errors:
                    import re
                    location = re.sub('Q{.+?}', '', location)

                severity = e.get('flag')
                if severity == 'warning' and not warnings:
                    continue

                errors.append({
                    'text': e.findtext('{*}text'),
                    'code': e.get('id'),
                    'severity': severity,
                    'test': e.get('test'),
                    'location': location,
                })

    return errors


def convert_schematron_file_to_xsl_file(input_sch_file, output_xsl_file, schxslt_path):
    # This function can be used to preprocess Schematron files. It
    # should only be used when there's a new release of the Schematron
    # file. schxscl_path is path to unzipped schxslt-*-xslt-only.zip -
    # the schxscl pipeline-for-svrl.xsl is used to convert the .sch
    # Schematron file to a regular .xls file.
    #
    # E.g. in a shell:
    # from utils.sendpeppol import convert_schematron_file_to_xsl_file
    # cd utils/sendpeppol-schematron
    # convert_schematron_file_to_xsl_file('PEPPOL-EN16931-UBL.sch', 'PEPPOL-EN16931-UBL.xsl', 'schxslt-1.9.5/')

    with open(input_sch_file, 'r') as f:
        input_text = f.read()

    with PySaxonProcessor(license=False) as proc:
        xsltproc = proc.new_xslt30_processor()
        document = proc.parse_xml(xml_text=input_text)
        executable = xsltproc.compile_stylesheet(stylesheet_file=os.path.join(schxslt_path, '2.0/pipeline-for-svrl.xsl'))
        output_text = executable.transform_to_string(xdm_node=document)

    with open(output_xsl_file, 'w') as f:
        f.write(output_text)



ENV_NS = 'http://www.w3.org/2003/05/soap-envelope'

WSS_BASE = 'http://docs.oasis-open.org/wss/2004/01/'
WSSE_NS = WSS_BASE + 'oasis-200401-wss-wssecurity-secext-1.0.xsd'
WSU_NS = WSS_BASE + 'oasis-200401-wss-wssecurity-utility-1.0.xsd'



class SendPeppolError(Exception):
    pass


def make_sendpeppol_error(output, code, temporary=False, **kwargs):
    e = SendPeppolError(output)
    e.code = code
    e.temporary = temporary
    for k, v in kwargs.items():
        setattr(e, k, v)
    return e



from typing import List

from lxml import etree
from lxml.builder import ElementMaker

from .exception import SendPeppolError, make_sendpeppol_error
from .validation import validate_peppol_document
from .sender import send_peppol_document, get_common_name_from_certificate

PEPPOL_ORGANIZATION_ID_TYPES = {
    # https://docs.peppol.eu/edelivery/codelists/v9.3/Peppol%20Code%20Lists%20-%20Participant%20identifier%20schemes%20v9.3.json
    # Backwards-compatibility additions
    'EAN': '0088',  # official name: GLN
    'DK:CVR': '0184',  # official name: DK:DIGST
    'NO:ORGNR': '0192',  # official name: NO:ORG
    'IS:KT': '0196',  # official name: IS:KTNR
    # Official list
    'FR:SIRENE': '0002',
    'SE:ORGNR': '0007',
    'FR:SIRET': '0009',
    'DUNS': '0060',
    'GLN': '0088',
    'DK:P': '0096',
    'IT:FTI': '0097',
    'NL:KVK': '0106',
    'EU:NAL': '0130',
    'IT:SIA': '0135',
    'IT:SECETI': '0142',
    'AU:ABN': '0151',
    'CH:UIDB': '0183',
    'DK:DIGST': '0184',
    'JP:SST': '0188',
    'NL:OINO': '0190',
    'EE:CC': '0191',
    'NO:ORG': '0192',
    'UBLBE': '0193',
    'SG:UEN': '0195',
    'IS:KTNR': '0196',
    'DK:ERST': '0198',
    'LEI': '0199',
    'LT:LEC': '0200',
    'IT:CUUO': '0201',
    'DE:LWID': '0204',
    'IT:COD': '0205',
    'BE:EN': '0208',
    'GS1': '0209',
    'IT:CFI': '0210',
    'IT:IVA': '0211',
    'FI:OVT2': '0216',
    'LV:URN': '0218',
    'JP:IIN': '0221',
    'FR:CTC': '0225',
    'MY:EIF': '0230',
    'AE:TIN': '0235',
    'LU:MAT': '0240',
    'SPIS': '0242',
    'HU:VAT': '9910',
    'EU:REID': '9913',
    'AT:VAT': '9914',
    'AT:GOV': '9915',
    'IBAN': '9918',
    'AT:KUR': '9919',
    'ES:VAT': '9920',
    'AD:VAT': '9922',
    'AL:VAT': '9923',
    'BA:VAT': '9924',
    'BE:VAT': '9925',
    'BG:VAT': '9926',
    'CH:VAT': '9927',
    'CY:VAT': '9928',
    'CZ:VAT': '9929',
    'DE:VAT': '9930',
    'EE:VAT': '9931',
    'GB:VAT': '9932',
    'GR:VAT': '9933',
    'HR:VAT': '9934',
    'IE:VAT': '9935',
    'LI:VAT': '9936',
    'LT:VAT': '9937',
    'LU:VAT': '9938',
    'LV:VAT': '9939',
    'MC:VAT': '9940',
    'ME:VAT': '9941',
    'MK:VAT': '9942',
    'MT:VAT': '9943',
    'NL:VAT': '9944',
    'PL:VAT': '9945',
    'PT:VAT': '9946',
    'RO:VAT': '9947',
    'RS:VAT': '9948',
    'SI:VAT': '9949',
    'SK:VAT': '9950',
    'SM:VAT': '9951',
    'TR:VAT': '9952',
    'VA:VAT': '9953',
    'FR:VAT': '9957',
    'US:EIN': '9959',
}
PEPPOL_REVERSED_ORGANIZATION_ID_TYPES = {v: k for k, v in PEPPOL_ORGANIZATION_ID_TYPES.items()}

PEPPOL_END_USER_STATISTICS_SCHEMATRON_XSLS = ['peppol-end-user-statistics-reporting-1.1.4.xsl']
PEPPOL_TRANSACTION_STATISTICS_SCHEMATRON_XSLS = ['peppol-transaction-statistics-reporting-1.0.4.xsl']

def skip_none_kwargs(kwargs):
    return {k: v for k, v in kwargs.items() if v is not None}

def clean_organization_id(organization_id, organization_id_type):
    if organization_id_type in PEPPOL_REVERSED_ORGANIZATION_ID_TYPES:
        scheme_id = organization_id_type
    else:
        scheme_id = PEPPOL_ORGANIZATION_ID_TYPES.get(organization_id_type)
    return scheme_id, organization_id

def generate_organization_id(E, ns, id_name, organization_id, organization_id_type):
    scheme_id, org_id = clean_organization_id(organization_id, organization_id_type)

    return E(ns("cbc", id_name), org_id, **skip_none_kwargs({'schemeID': scheme_id}))

def generate_common_name(certfile):
    with open(certfile, 'rb') as sender_certfile_f:
        sender_cert = sender_certfile_f.read()
    return get_common_name_from_certificate(sender_cert)

def generate_peppol_statistics_header(E, from_date, to_date, sender_common_name):
    return E('Header',
             E('ReportPeriod',
               E('StartDate', from_date.isoformat()),
               E('EndDate', to_date.isoformat())),
             E('ReporterID', sender_common_name, schemeID='CertSubjectCN'))

def render_peppol_end_user_statistics_xml(stats, sender_common_name):
    E = ElementMaker()

    xml = E('EndUserStatisticsReport',
            E('CustomizationID', 'urn:fdc:peppol.eu:edec:trns:end-user-statistics-report:1.1'),
            E('ProfileID', 'urn:fdc:peppol.eu:edec:bis:reporting:1.0'),
            generate_peppol_statistics_header(E, stats['from_date'], stats['to_date'], sender_common_name),
            xmlns='urn:fdc:peppol:end-user-statistics-report:1.1')

    def generate_senders_and_receivers(s, r=None):
        if s is None:
            s = set()
        if r is None:
            r = set()
        return [E('SendingEndUsers', str(len(s))),
                E('ReceivingEndUsers', str(len(r))),
                E('SendingOrReceivingEndUsers', str(len(s | r)))]

    xml.append(E('FullSet',
                 *generate_senders_and_receivers(stats['senders'])
                 ))

    for country_code, senders in stats['senders_by_country'].items():
        xml.append(E('Subset',
                     E('Key', country_code, metaSchemeID="CC", schemeID="EndUserCountry"),
                     *generate_senders_and_receivers(senders),
                     type='PerEUC'))

    for (document_type, country_code), senders in stats['senders_by_document_type_country'].items():
        xml.append(E('Subset',
                     E('Key', document_type, metaSchemeID="DT", schemeID="busdox-docid-qns"),
                     E('Key', country_code, metaSchemeID="CC", schemeID="EndUserCountry"),
                     *generate_senders_and_receivers(senders),
                     type='PerDT-EUC'))

    for (document_type, process_type), senders in stats['senders_by_document_type_process_type'].items():
        xml.append(E('Subset',
                     E('Key', document_type, metaSchemeID="DT", schemeID="busdox-docid-qns"),
                     E('Key', process_type, metaSchemeID="PR", schemeID="cenbii-procid-ubl"),
                     *generate_senders_and_receivers(senders),
                     type='PerDT-PR'))

    for (document_type, process_type, country_code), senders in stats['senders_by_document_type_process_type_country'].items():
        xml.append(E('Subset',
                     E('Key', document_type, metaSchemeID="DT", schemeID="busdox-docid-qns"),
                     E('Key', process_type, metaSchemeID="PR", schemeID="cenbii-procid-ubl"),
                     E('Key', country_code, metaSchemeID="CC", schemeID="EndUserCountry"),
                     *generate_senders_and_receivers(senders),
                     type='PerDT-PR-EUC'))

    return etree.tostring(xml, xml_declaration=True, encoding='UTF-8', method='xml')


def render_peppol_transaction_statistics_xml(stats, certfile):
    E = ElementMaker()

    xml = E('TransactionStatisticsReport',
            E('CustomizationID', 'urn:fdc:peppol.eu:edec:trns:transaction-statistics-reporting:1.0'),
            E('ProfileID', 'urn:fdc:peppol.eu:edec:bis:reporting:1.0'),
            generate_peppol_statistics_header(E, stats['from_date'], stats['to_date'], certfile),
            xmlns='urn:fdc:peppol:transaction-statistics-report:1.0')

    xml.append(E('Total',
                 E('Incoming', '0'),
                 E('Outgoing', str(stats['outgoing']))
                 ))

    for transport_profile, outgoing in stats['outgoing_by_transport_profile'].items():
        xml.append(E('Subtotal',
                     E('Key', transport_profile, metaSchemeID="TP", schemeID="Peppol"),
                     E('Incoming', '0'),
                     E('Outgoing', str(outgoing)),
                     type='PerTP'))

    for (receiver_common_name, document_type, process_type), outgoing in stats['outgoing_by_receiver_common_name_document_type_process_type'].items():
        xml.append(E('Subtotal',
                     E('Key', receiver_common_name, metaSchemeID="SP", schemeID="CertSubjectCN"),
                     E('Key', document_type, metaSchemeID="DT", schemeID="busdox-docid-qns"),
                     E('Key', process_type, metaSchemeID="PR", schemeID="cenbii-procid-ubl"),
                     E('Incoming', '0'),
                     E('Outgoing', str(outgoing)),
                     type='PerSP-DT-PR'))

    # we don't output PerSP-DT-CC as we're only sending

    return etree.tostring(xml, xml_declaration=True, encoding='UTF-8', method='xml')

def send_peppol_statistics(
    aggr_stats: dict,
    our_endpoint: dict,
    xmlsec_path: str,
    keyfile: str,
    password: str,
    certfile: str,
    test_environment: bool,
    receiver_id: str='9925:BE0848934496'
) -> List[dict]:
    """
    Send peppol statistics to the required reporting endpoint.

    ``aggr_stats`` (dict) should look like this:
    ```
    document_type = 'urn:oasis:names:specification:ubl:schema:xsd:Invoice-2::Invoice##urn:cen.eu:en16931:2017#compliant#urn:fdc:peppol.eu:2017:poacc:billing:3.0::2.1',
    process_type = 'urn:fdc:peppol.eu:2017:poacc:billing:01:1.0'
    transport_profile = 'peppol-transport-as4-v2_0'
    sender_country = 'DK'
    receiver_common_name = 'PNO000063'
    aggr_stats = {
      'from_date': <DATETIME>,
      'to_date': <DATETIME>,
      'outgoing': <NUM>,
      'outgoing_by_transport_profile': { transport_profile: <NUM> },
      'outgoing_by_receiver_common_name_document_type_process_type': { (receiver_common_name, document_type, process_type): <NUM> },
      'senders': <SET_OF_IDS>,
      'senders_by_country': { sender_country: <SET_OF_IDS> },
      'senders_by_document_type_country': { (document_type, country): <SET_OF_IDS> },
      'senders_by_document_type_process_type': { (document_type, process_type): <SET_OF_IDS> },
      'senders_by_document_type_process_type_country': { (document_type, process_type, sender_country): <SET_OF_IDS> }
    }
    ```

    ``our_endpoint`` (dict) should look like this: ``{ "id": "PDK000592", "type": "DK:P", "country": "DK" }``

    ``xmlsec_path`` (str): specifies the path to a xmlsec 1.3 or higher binary.

    ``keyfile`` (str): the path to the private key of the sender.

    ``password`` (str): the password for the private key of the sender.

    ``certfile`` (str): the path to the public key of the sender.

    ``test_environment`` (bool): use test SML servers?

    ``receiver_id`` (str): Receiver participant, defaults to the OpenPeppol organization
    """
    sender_common_name = generate_common_name(certfile)
    end_user_xml = render_peppol_end_user_statistics_xml(aggr_stats, sender_common_name)
    transaction_xml = render_peppol_transaction_statistics_xml(aggr_stats, sender_common_name)

    sender_id_element = generate_organization_id(ElementMaker(), lambda shorthand, tag: tag, "EndpointID", our_endpoint['id'], our_endpoint['type'])
    sender_id = sender_id_element.get('schemeID') + ':' + sender_id_element.text

    results = []
    failure = None
    for xml, schematron_xsls in [(end_user_xml, PEPPOL_END_USER_STATISTICS_SCHEMATRON_XSLS), (transaction_xml, PEPPOL_TRANSACTION_STATISTICS_SCHEMATRON_XSLS)]:
        errors = validate_peppol_document(xml, schematron_xsls)
        if errors:
            for d in errors:
                print(d)
            raise make_sendpeppol_error('Schematron validation failed.', 'validation', False)

        try:
            results.append(
                send_peppol_document(
                    xml, xmlsec_path, keyfile, password, certfile, sender_id=sender_id, receiver_id=receiver_id, sender_country=our_endpoint['country'], test_environment=test_environment, timeout=20,
                )
            )
        except SendPeppolError as ex:
            print(f"Failed with: {ex.code} {ex}")
            failure = ex
    if failure:
        raise failure
    return results


from .exception import SendPeppolError
from .sender import send_peppol_document
from .statistics import send_peppol_statistics
from .validation import validate_peppol_document

__all__ = [
    "SendPeppolError",
    "send_peppol_document",
    "send_peppol_statistics",
    "validate_peppol_document",
]



import uuid
import socket

from .xmlhelpers import get_element_maker
from .constants import ENV_NS, WSSE_NS, WSU_NS

NS2_NS = 'http://docs.oasis-open.org/ebxml-msg/ebms/v3.0/ns/core/200704/'
STBH_NS = 'http://www.unece.org/cefact/namespaces/StandardBusinessDocumentHeader'

NAMESPACES = {
    'env': ENV_NS,
    'ns2': NS2_NS,
    'wsu': WSU_NS,
    'wsse': WSSE_NS,
    'stbh': STBH_NS
}

def generate_as4_envelope(utc_timestamp, document_type, process_type, sender_id, receiver_id, to_party_id, service_provider_id):
    E, ns = get_element_maker(NAMESPACES)

    hostname = socket.getfqdn()

    attachment_id = 'cid:{}@{}'.format(uuid.uuid4(), hostname)
    message_id = '{}@{}'.format(uuid.uuid4(), hostname)

    user_message = E(ns('ns2', 'UserMessage'),
                    E(ns('ns2', 'MessageInfo'),
                      E(ns('ns2', 'Timestamp'), utc_timestamp),
                      E(ns('ns2', 'MessageId'), message_id)),
                    E(ns('ns2', 'PartyInfo'),
                      E(ns('ns2', 'From'),
                        E(ns('ns2', 'PartyId'), service_provider_id, type='urn:fdc:peppol.eu:2017:identifiers:ap'),
                        E(ns('ns2', 'Role'), 'http://docs.oasis-open.org/ebxml-msg/ebms/v3.0/ns/core/200704/initiator')),
                      E(ns('ns2', 'To'),
                        E(ns('ns2', 'PartyId'), to_party_id, type='urn:fdc:peppol.eu:2017:identifiers:ap'),
                        E(ns('ns2', 'Role'), 'http://docs.oasis-open.org/ebxml-msg/ebms/v3.0/ns/core/200704/responder'))),
                    E(ns('ns2', 'CollaborationInfo'),
                      E(ns('ns2', 'AgreementRef'), 'urn:fdc:peppol.eu:2017:agreements:tia:ap_provider'),
                      E(ns('ns2', 'Service'), process_type, type='cenbii-procid-ubl'),
                      E(ns('ns2', 'Action'), 'busdox-docid-qns::{}'.format(document_type)),
                      E(ns('ns2', 'ConversationId'), message_id)),
                    E(ns('ns2', 'MessageProperties'),
                      E(ns('ns2', 'Property'), sender_id, name='originalSender', type='iso6523-actorid-upis'),
                      E(ns('ns2', 'Property'), receiver_id, name='finalRecipient', type='iso6523-actorid-upis')),
                    E(ns('ns2', 'PayloadInfo'),
                      E(ns('ns2', 'PartInfo'),
                        E(ns('ns2', 'PartProperties'),
                          E(ns('ns2', 'Property'), 'application/gzip', name='CompressionType'),
                          E(ns('ns2', 'Property'), 'application/xml', name='MimeType')), href=attachment_id)))

    messaging = E(ns('ns2', 'Messaging'), user_message, {ns('env', 'mustUnderstand'): 'true', ns('wsu', 'Id'): 'messaging'})

    body = E(ns('env', 'Body'), {ns('wsu', 'Id'): 'body'})

    envelope = E(ns('env', 'Envelope'),
                 E(ns('env', 'Header'),
                   messaging,
                   E(ns('wsse', 'Security'), { ns('env', 'mustUnderstand'): 'true' })),
                 body)

    return attachment_id, envelope, messaging, body



import datetime
from pathlib import Path

from lxml import etree
from lxml.builder import ElementMaker

from .smp import get_service_info_for_participant_from_smp
from .exception import make_sendpeppol_error
from .as4_sender import get_headers_and_body_for_posting_as4_document
from .as4_sender import post_edelivery_as4_document

def split_tag(element):
    ns, tag = element.tag.lstrip('{').split('}')
    return ns, tag

def get_document_type_from_ubl(ubl, document_type_version=None):
    root_ns, root_tag = split_tag(ubl)

    if document_type_version is None:
        document_type_version = ubl.findtext('./{*}CustomizationID').rsplit(':', 1)[1]

    return root_ns + '::' + root_tag + '##' + ubl.findtext('./{*}CustomizationID') + '::' + document_type_version

def wrap_ubl_in_peppol_standard_business_document_header(ubl, utc_timestamp, document_type, process_type, sender_id, sender_country, receiver_id):
    E = ElementMaker(nsmap={})

    unique_id = utc_timestamp.replace(':', '')

    id_element = ubl.find('./{*}ID')
    if id_element is not None:
        unique_id = id_element.text + '-' + unique_id

    root_ns, root_tag = split_tag(ubl)

    document_type_version = document_type.rsplit(':', 1)[1]

    header = E("StandardBusinessDocumentHeader",
          E("HeaderVersion", "1.0"),
          E("Sender",
            E("Identifier", sender_id, Authority="iso6523-actorid-upis")),
          E("Receiver",
            E("Identifier", receiver_id, Authority="iso6523-actorid-upis")),
          E("DocumentIdentification",
            E("Standard", root_ns),
            E("TypeVersion", document_type_version),
            E("InstanceIdentifier", unique_id),
            E("Type", root_tag),
            E("CreationDateAndTime", utc_timestamp)
            ),
          E("BusinessScope",
            E("Scope",
              E("Type", "DOCUMENTID"),
              E("InstanceIdentifier", document_type),
              E("Identifier", "busdox-docid-qns")
              ),
            E("Scope",
              E("Type", "PROCESSID"),
              E("InstanceIdentifier", process_type),
              E("Identifier", "cenbii-procid-ubl")
              ),
            E("Scope",
              E("Type", "COUNTRY_C1"),
              E("InstanceIdentifier", sender_country),
              ),
            )
          )

    business_doc = E("StandardBusinessDocument", header, ubl, xmlns="http://www.unece.org/cefact/namespaces/StandardBusinessDocumentHeader")

    return etree.tostring(business_doc), business_doc

def get_common_name_from_certificate(cert):
    import asn1crypto.pem
    import asn1crypto.x509

    cert_type_name, cert_headers, cert_der_bytes = asn1crypto.pem.unarmor(cert)
    assert cert_type_name == 'CERTIFICATE'
    parsed_cert = asn1crypto.x509.Certificate.load(cert_der_bytes)

    return parsed_cert.subject.native['common_name']

def validate_certificate(cert, test):
    import asn1crypto.pem
    from certvalidator import ValidationContext, CertificateValidator

    certfolder = Path(__file__).parent / "data" / "certs"
    if datetime.datetime.now(datetime.UTC) > datetime.datetime(2026, 4, 1, 0, 0, 0, tzinfo=datetime.UTC):
        # See G2→G3 migration plan: https://openpeppol.atlassian.net/wiki/spaces/RR/pages/4387602465/2025.06.17+PKI+Migration+Plan
        cabundle = certfolder / ('ap-test-truststore-g3.pem' if test else 'ap-prod-truststore-g3.pem')
    else:
        cabundle = certfolder / ('ap-test-truststore-g2-g3.pem' if test else 'ap-prod-truststore-g2-g3.pem')
    trust_roots = []
    with open(cabundle, 'rb') as f:
        for _, _, der_bytes in asn1crypto.pem.unarmor(f.read(), multiple=True):
            trust_roots.append(der_bytes)

    context = ValidationContext(
        allow_fetching=True,
        trust_roots=trust_roots,
        revocation_mode="hard-fail",
    )
    validator = CertificateValidator(cert, validation_context=context)
    validator.validate_usage(
        key_usage={"key_encipherment", "key_agreement", "digital_signature"},
        extended_key_usage={"client_auth"},
        extended_optional=False,
    )


def send_peppol_document(
    document_content: bytes,
    xmlsec_path: str,
    keyfile: str,
    keyfile_password: str,
    certfile: str,
    sender_id: str=None,
    receiver_id: str=None,
    sender_country: str=None,
    document_type_version: str=None,
    test_environment: bool=True,
    timeout: int=20,
    dryrun: bool=False,
) -> dict:
    """
    Send a peppol document. Returned is a dictionary of information you need to record to later send reports to Peppol.

    ``document_content`` (bytes): document to send. Note the standard business header will automatically be added.

    ``xmlsec_path`` (str): specifies the path to a xmlsec 1.3 or higher binary.

    ``keyfile`` (str): the path to the private key of the sender.

    ``password`` (str): the password for the private key of the sender.

    ``certfile`` (str): the path to the public key of the sender.

    ``service_provider_id`` (bool): identifier of the sending service provider.

    ``sender_id`` (str): optional sender id, will be extracted from document if not specified.

    ``receiver_id`` (str): optional receiver id, will be extracted from document if not specified.

    ``sender_country`` (str): optional sender country, will be extracted from document if not specified.

    ``document_type_version`` (str): the document type version, if not specified will be last part of CustomizationID.
                                     For invoices should be set to `2.1`.

    ``test_environment`` (bool): use test SML servers?

    ``timeout`` (int): number of seconds to wait for response from the remote end.

    ``dryrun`` (bool): if specified, will prepare, get the endpoint, test document for validation errors but not send to remote endpoint.
                       Return value will be a tuple of ``body, header, stats``.
    """
    document_xml = etree.fromstring(document_content)
    #print(etree.tostring(document_xml, pretty_print=True).decode())

    if sender_id is None:
        element = document_xml.find("./{*}AccountingSupplierParty/{*}Party/{*}EndpointID")
        sender_id = element.attrib.get('schemeID', '') + ':' + element.text
    if receiver_id is None:
        element = document_xml.find("./{*}AccountingCustomerParty/{*}Party/{*}EndpointID")
        receiver_id = element.attrib.get('schemeID', '') + ':' + element.text
    if sender_country is None:
        sender_country = document_xml.findtext('./{*}AccountingSupplierParty/{*}Party/{*}PostalAddress/{*}Country/{*}IdentificationCode')

    sender_end_user = None
    element = document_xml.find('./{*}AccountingSupplierParty/{*}Party/{*}PartyIdentification/{*}ID')
    if element is not None:
        sender_end_user = element.attrib.get('schemeID', '') + ':' + element.text

    document_type = get_document_type_from_ubl(document_xml, document_type_version=document_type_version)
    process_type = document_xml.findtext('./{*}ProfileID')

    utc_timestamp = datetime.datetime.now(tz=datetime.timezone.utc).replace(microsecond=0).isoformat()

    document_content, document_xml = wrap_ubl_in_peppol_standard_business_document_header(document_xml, utc_timestamp, document_type, process_type, sender_id, sender_country, receiver_id)

    transport_profile, endpoint_url, receiver_cert = get_service_info_for_participant_from_smp(receiver_id, document_type, test_environment=test_environment, timeout=timeout)

    if not endpoint_url:
        raise make_sendpeppol_error("Endpoint URL not found", 'not-found-in-smp')

    if not receiver_cert:
        raise make_sendpeppol_error("Receiver certificate not found", 'not-found-in-smp')

    with open(certfile, 'rb') as sender_certfile_f:
        sender_cert = sender_certfile_f.read()
    service_provider_id = get_common_name_from_certificate(sender_cert)

    validate_certificate(receiver_cert, test_environment)
    to_party_id = get_common_name_from_certificate(receiver_cert)

    stats = {
        'timestamp': utc_timestamp,
        'receiver_common_name': to_party_id,
        'document_type': document_type,
        'process_type': process_type,
        'transport_profile': transport_profile,
        'sender_end_user': sender_end_user,
        'sender_country': sender_country,
    }

    body, headers = get_headers_and_body_for_posting_as4_document(document_content, document_xml, utc_timestamp, document_type, process_type, sender_id, receiver_id, to_party_id, xmlsec_path, keyfile, keyfile_password, sender_cert, receiver_cert, service_provider_id)

    if dryrun:
        return body, headers, stats

    post_edelivery_as4_document(endpoint_url, body, headers, timeout=timeout)

    return stats



"""
Functions for WS-Security (WSSE) signing + encrypting
"""

import base64
import io
from pathlib import Path

from lxml import etree
import xmlsec
import xml.etree.ElementTree
import hashlib
import gzip

from .xmlhelpers import get_element_maker
from .constants import ENV_NS, WSSE_NS, WSU_NS, WSS_BASE
from .exception import make_sendpeppol_error

DS_NS = 'http://www.w3.org/2000/09/xmldsig#'
ENC_NS = 'http://www.w3.org/2001/04/xmlenc#'
ENC11_NS = 'http://www.w3.org/2009/xmlenc11#'
WSSE11_NS = 'http://docs.oasis-open.org/wss/oasis-wss-wssecurity-secext-1.1.xsd'

XML_CANONICAL_C14N = 'http://www.w3.org/2001/10/xml-exc-c14n#'
WSS_ATTACHMENT_SIGNATURE_TRANSFORM = 'http://docs.oasis-open.org/wss/oasis-wss-SwAProfile-1.1#Attachment-Content-Signature-Transform'

WSS_BASE64B_ENCODING_TYPE = WSS_BASE + 'oasis-200401-wss-soap-message-security-1.0#Base64Binary'
WSS_X509TOKEN = WSS_BASE + 'oasis-200401-wss-x509-token-profile-1.0#X509v3'

XML_SHA256 = 'http://www.w3.org/2001/04/xmlenc#sha256'
XML_RSA_SHA256 = 'http://www.w3.org/2001/04/xmldsig-more#rsa-sha256'
XML_RSA_OAEP = 'http://www.w3.org/2009/xmlenc11#rsa-oaep'
XML_MGF_SHA256 = 'http://www.w3.org/2009/xmlenc11#mgf1sha256'

NAMESPACES = {
    'env': ENV_NS,
    'wsu': WSU_NS,
    'wsse': WSSE_NS,
    'wsse11': WSSE11_NS,
    'ec': XML_CANONICAL_C14N,
    'xenc': ENC_NS,
    'xenc11': ENC11_NS,
    'ds': DS_NS
}

# We need to manually "craft" a signature because xmlsec doesn't
# properly handle external content. It does support cid and tries to
# resolve it to a file in the current working directory, but this only
# works if the file referenced is an xml file. This standard makes no
# sense, why would you use a "canonical" (c14n) encoding before
# gzipping, if gzip itself is implementation specific?
def sign_as4_envelope_and_body(envelope, attachment_id, doc_hash, body, messaging, cert, keyfile, password):
    E, ns = get_element_maker(NAMESPACES)
    header = envelope.find(ns("env", 'Header'))
    security = header.find(ns("wsse", 'Security'))

    security_token = generate_binary_security_token_xml(E, ns, cert, 'signkey')
    security.insert(0, security_token)

    # definition of "canonical":
    # - have all namespaces so that it is valid xml (even if this is not what is submitted over the network)
    # - be transformed using c14n (NOT c14n2!) with the exclude option
    # - proper indentation depending on where in the tree the element is situated
    def canonical_as4_xml(element, indentation_level=0):
        out = io.BytesIO()
        xml.etree.ElementTree.indent(element, ' ', indentation_level)
        etree.ElementTree(element).write(out, method='c14n', exclusive=True)
        return out.getvalue()

    def generate_xml_hash(element, indentation_level=0):
        return base64.b64encode(hashlib.sha256(canonical_as4_xml(element, indentation_level)).digest()).decode('utf-8')

    messaging_hash = generate_xml_hash(messaging, 5)
    messaging_id = messaging.get(ns("wsu", "Id"))

    body_hash = generate_xml_hash(body)
    body_id = body.get(ns("wsu", "Id"))

    signature_info = E(ns('ds', 'SignedInfo'),
                       E(ns('ds', 'CanonicalizationMethod'),
                         E(ns('ec', 'InclusiveNamespaces'), PrefixList='env'),
                         Algorithm=XML_CANONICAL_C14N),
                       E(ns('ds', 'SignatureMethod'), Algorithm=XML_RSA_SHA256),
                       generate_digest_reference(E, ns, body_id, XML_CANONICAL_C14N, body_hash),
                       generate_digest_reference(E, ns, messaging_id, XML_CANONICAL_C14N, messaging_hash),
                       generate_digest_reference(E, ns, attachment_id, WSS_ATTACHMENT_SIGNATURE_TRANSFORM, doc_hash))

    key = xmlsec.Key.from_file(keyfile, xmlsec.KeyFormat.PEM, password)
    key.load_cert_from_memory(cert, xmlsec.KeyFormat.PEM)

    ctx = xmlsec.SignatureContext()
    ctx.key = key

    signature_info_str = canonical_as4_xml(signature_info).decode('utf-8')
    # why is env namespace missing?
    signature_info_str = signature_info_str.replace(
        '<ds:SignedInfo xmlns:ds="http://www.w3.org/2000/09/xmldsig#">',
        f'<ds:SignedInfo xmlns:ds="http://www.w3.org/2000/09/xmldsig#" xmlns:env="{ENV_NS}">')

    sign = ctx.sign_binary(signature_info_str, xmlsec.constants.TransformRsaSha256)
    signature_value = base64.b64encode(sign).decode('utf-8')

    key_info = generate_key_info_with_security_token(E, ns, security_token)

    security.insert(1, E(ns("ds", "Signature"),
                         signature_info,
                         E(ns("ds", "SignatureValue"), signature_value),
                         key_info))

def insert_encryption_info_in_as4_envelope(envelope, receiver_cert, cipher_value, attachment_id):
    E, ns = get_element_maker(NAMESPACES)
    header = envelope.find(ns("env", 'Header'))
    security_node = header.find(ns("wsse", 'Security'))

    security_token = generate_binary_security_token_xml(E, ns, receiver_cert, 'encryptkey')
    key_info = generate_key_info_with_security_token(E, ns, security_token)

    encrypted_key = E(ns('xenc', 'EncryptedKey'),
                      E(ns('xenc', 'EncryptionMethod'),
                        E(ns('ds', 'DigestMethod'), Algorithm=XML_SHA256),
                        E(ns('xenc11', 'MGF'), Algorithm=XML_MGF_SHA256),
                        Algorithm=XML_RSA_OAEP),
                      key_info,
                      E(ns('xenc', 'CipherData'),
                        E(ns('xenc', 'CipherValue'), cipher_value)),
                      E(ns('xenc', 'ReferenceList'),
                        E(ns('xenc', 'DataReference'), URI='#encrypteddata')),
                      Id='encryptedkey')

    encrypted_data = E(ns('xenc', 'EncryptedData'),
                       E(ns('xenc', 'EncryptionMethod'), Algorithm='http://www.w3.org/2009/xmlenc11#aes128-gcm'),
                       E(ns('ds', 'KeyInfo'),
                         E(ns('wsse', 'SecurityTokenReference'),
                           E(ns('wsse', 'Reference'), URI='#encryptedkey'),
                           { ns('wsse11', 'TokenType'):
                             'http://docs.oasis-open.org/wss/oasis-wss-soap-message-security-1.1#EncryptedKey' }
                           )),
                       E(ns('xenc', 'CipherData'),
                         E(ns('xenc', 'CipherReference'),
                           E(ns('xenc', 'Transforms'),
                             E(ns('ds', 'Transform'),
                               Algorithm='http://docs.oasis-open.org/wss/oasis-wss-SwAProfile-1.1#Attachment-Ciphertext-Transform')
                             ),
                           URI=attachment_id)),
                       Id='encrypteddata', MimeType='application/octet-stream',
                       Type='http://docs.oasis-open.org/wss/oasis-wss-SwAProfile-1.1#Attachment-Content-Only')

    security_node.insert(0, security_token)
    security_node.insert(1, encrypted_key)
    security_node.insert(2, encrypted_data)

def encrypt_as4_document(document_content, receiver_cert, xmlsec_path):
    E, ns = get_element_maker(NAMESPACES)

    gzipped_content = gzip.compress(document_content, compresslevel=6)
    gzipped_document_hash = base64.b64encode(hashlib.sha256(gzipped_content).digest()).decode('ascii')

    import subprocess
    import tempfile

    with tempfile.NamedTemporaryFile(prefix='temp-sendpeppol-certfile-', suffix='.pem') as receiver_certfile:
        receiver_certfile.write(receiver_cert)
        receiver_certfile.flush()

        # FIXME: it would be better to use the Python wrapper so we could
        # avoid all this file mangling, but it doesn't support xmlsec 1.3
        # yet
        xmlsec_args = [
            xmlsec_path, '--encrypt',
            '--pubkey-cert-pem', receiver_certfile.file.name,
            '--session-key', 'aes-128',
            '--binary-data', '/dev/stdin',
            #'--verbose'
            '--lax-key-search',
            Path(__file__).parent / 'data' / 'xmlsec' / 'sendpeppol_encrypt_template.xml',
        ]

        xmlsec_popen = subprocess.Popen(
            xmlsec_args,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        xmlsec_output, xmlsec_errors = xmlsec_popen.communicate(input=gzipped_content, timeout=20)
        if xmlsec_popen.returncode != 0:
            raise make_sendpeppol_error('xmlsec: ' + xmlsec_errors.decode(), 'encryption')

    xmlsec_xml = etree.fromstring(xmlsec_output)

    cipher_values = [a.text for a in xmlsec_xml.iter() if a.tag == ns("xenc", 'CipherValue')]
    cipher_value = cipher_values[0].replace('\n', '')
    encrypted_gzip_b64 = cipher_values[1].replace('\n', '')
    encrypted_gzip = base64.b64decode(encrypted_gzip_b64.encode('ascii'))

    return cipher_value, encrypted_gzip, gzipped_document_hash

### HELPERS ###

def generate_digest_reference(E, ns, ref_id, transform, digest_value):
    if transform != WSS_ATTACHMENT_SIGNATURE_TRANSFORM:
        ref_id = '#' + ref_id

    return E(ns('ds', 'Reference'),
             E(ns('ds', 'Transforms'),
               E(ns('ds', 'Transform'), Algorithm=transform)),
             E(ns('ds', 'DigestMethod'), Algorithm=XML_SHA256),
             E(ns('ds', 'DigestValue'), digest_value),
             URI=ref_id)

def generate_key_info_with_security_token(E, ns, security_token):
    return E(ns('ds', 'KeyInfo'),
             E(ns('wsse', 'SecurityTokenReference'),
               E(ns('wsse', 'Reference'),
                 ValueType=security_token.get('ValueType'),
                 URI=f"#{security_token.get(ns('wsu', 'Id'))}"),
               { ns('wsse', 'TokenType'): security_token.get('ValueType') }))

def generate_binary_security_token_xml(E, ns, cert, identifier):
    import asn1crypto.pem

    type_name, headers, der_bytes = asn1crypto.pem.unarmor(cert)
    b64_cert = base64.b64encode(der_bytes).decode('utf-8')

    return E(ns('wsse', 'BinarySecurityToken'), b64_cert,
             {ns('wsu', 'Id'): identifier, 'EncodingType': WSS_BASE64B_ENCODING_TYPE, 'ValueType': WSS_X509TOKEN})



import re

from lxml import etree

import base64
import dns.resolver
import urllib.parse
import requests
import hashlib

from .exception import make_sendpeppol_error


def get_smp_url_from_dns(participant_id, test_environment):
    smp_id = base64.b32encode(hashlib.sha256(participant_id.lower().encode()).digest()).decode().rstrip("=")

    if test_environment:
        base_hostname = 'acc.edelivery.tech.ec.europa.eu'
    else:
        base_hostname = 'edelivery.tech.ec.europa.eu'

    smp_domain = f'{smp_id}.iso6523-actorid-upis.{base_hostname}'

    resolver = dns.resolver.Resolver()
    try:
        answers = resolver.resolve(smp_domain, 'NAPTR')
    except (dns.resolver.NXDOMAIN, dns.resolver.NoAnswer) as e:
        raise make_sendpeppol_error(str(e), 'dns-error', temporary=False)
    except dns.resolver.DNSException as e:
        raise make_sendpeppol_error(str(e), 'dns-error', temporary=True)

    result = None
    for rdata in sorted(answers, key=lambda rdata: rdata.order):
        # This is not a full implementation of the NAPTR spec (RFC2915) but should support everything used by Peppol
        # and be somewhat safe against unexpected record contents.
        if rdata.flags.decode().lower() == "u" and rdata.service.decode().lower() == "meta:smp":
            delim_char = rdata.regexp.decode()[0]
            regex_data = rdata.regexp.decode().split(delim_char)
            result = re.sub('^' + regex_data[1] + '$', regex_data[2], smp_domain)

            if not result.startswith("https://") and not result.startswith("http://"):
                raise make_sendpeppol_error("Regexp result of NATPR record is not an URL", 'dns-error', temporary=False)
            break

    if not result:
        raise make_sendpeppol_error("The NATPR record could not be resolved to an URL", 'dns-error', temporary=False)

    # SMP: domain + path -> xml with service descriptions
    # get all available interfaces (invoice, credit note etc.)
    return result + "/iso6523-actorid-upis::" + participant_id


def get_service_urls_for_participant_from_smp(participant_id, test_environment, timeout):
    # SML: receiver -> SMP domain (DNS)
    smp_url = get_smp_url_from_dns(participant_id, test_environment)

    try:
        r = requests.get(smp_url, timeout=timeout)
        r.raise_for_status()
    except (ConnectionError, requests.exceptions.RequestException) as e:
        temporary = True
        code = 'server-error'
        if isinstance(e, requests.exceptions.HTTPError) and e.response and e.response.status_code < 500:
            temporary = False
            code = 'not-found-in-smp'

        raise make_sendpeppol_error(str(e), code, temporary=temporary, url=smp_url)

    try:
        service_urls_xml = etree.fromstring(r.content)
    except ValueError as e:
        raise make_sendpeppol_error(str(e), 'server-error', temporary=True, url=smp_url)

    service_urls = []
    for el in service_urls_xml.findall(".//{*}ServiceMetadataReference"):
        href = el.get('href')
        if href:
            service_urls.append(href)

    return service_urls


def get_service_info_for_participant_from_smp(participant_id, document_type, test_environment, timeout):
    """Lookup participant in Peppol Service Metadata Publisher for
    information on service type. Service type is Scope.Identifier +
    '::' + Scope.InstanceIdentifier from the Peppol header."""

    service_urls = get_service_urls_for_participant_from_smp(participant_id, test_environment, timeout)
    service_url = next((url for url in service_urls if url.endswith(urllib.parse.quote(document_type))), None)
    if not service_url:
        raise make_sendpeppol_error("{0} not found in {1}".format(document_type, [urllib.parse.unquote(url) for url in service_urls]), 'not-found-in-smp')

    try:
        r = requests.get(service_url, timeout=timeout)
        r.raise_for_status()
    except (ConnectionError, requests.exceptions.RequestException) as e:
        temporary = True
        code = 'server-error'
        if isinstance(e, requests.exceptions.HTTPError) and e.response and e.response.status_code < 500:
            temporary = False
            code = 'not-found-in-smp'

        raise make_sendpeppol_error(str(e), code, temporary=temporary, url=service_url)

    try:
        service_info_xml = etree.fromstring(r.content)
    except ValueError as e:
        raise make_sendpeppol_error(str(e), 'server-error', temporary=True)

    #print(r.content.decode().replace('>', '>\n'))

    # we don't check process match at the moment - to do that we'd to
    # get the process identifier from the Peppol header
    # <Type>PROCESSID</type> like we get the service identifier
    transport_profile = endpoint_url = certificate = None
    for endpoint_e in service_info_xml.findall('.//{*}ServiceInformation//{*}Process//{*}Endpoint'):
        transport_profile = endpoint_e.get('transportProfile')

        if not transport_profile.startswith('peppol-transport-as4'):
            continue

        endpoint_url = endpoint_e.findtext('./{*}EndpointReference/{*}Address')
        certificate = endpoint_e.findtext('./{*}Certificate')
        if certificate:
            certificate = b'-----BEGIN CERTIFICATE-----\n' + certificate.encode('ascii') + b'\n-----END CERTIFICATE-----'

        break

    return transport_profile, endpoint_url, certificate
